#include "GCCIM.hpp"
#include "DataBases.h"

typedef std::vector < int > VECELE;
typedef std::vector < VECELE > MATELE;

using namespace std;
using namespace Gecode;

	///---- Variables ----///
	vector < vector < int > > Solutions; // to store closed patterns
	vector <int> Frequence_Solution; // to store frequency of each closed pattern
	int CurFreq = 1;
	vector <int> Code; // item's encoding
	vector< boost::dynamic_bitset<> > Vertical_DataBase;
	vector< boost::dynamic_bitset<> > Horizontal_DataBase;
	bool print_ = true; // allow print of encoding ,solutions and database under different representations
	
	Cover coverD; // instance of Cover class
	DataBases DataBasesInstance; // instance of DataBases class	
	Search::Statistics GecodeStat; // Gecode's Statistics
					
	///---- Fonction's prototypes ----///
	void Resolution(int MINFREQ, int MINSIZE); // Resolution	
	void printStatsGecode(const Gecode::Search::Statistics &stats); // # explored nodes, tree depth ...	
	void PostTraitement(vector < vector <int> > itemset, vector <int> Code, vector <int> Frequence_Solution); // post-processing of solutions according to encoding => "Code"
			
	///---- Model -----///
	class ItemSetClosed: public Script {	
		protected:
		
		IntVarArray Itemset; // pattern we are looking for
		//IntVar Size_Itemset;  // pattern size		
				
		public:	
		ItemSetClosed(const Options& opt, int MINFREQ, int MINSIZE): /* Script(opt),pour que le code tourne sous 4.4.0*/ Itemset(*this, Code.size(),0, 1)/*, Size_Itemset(*this, 0, Code.size())*/{
			
 			if(MINSIZE > 0){ // if the user defines a minimal size (default = 0)
				//count(*this, Itemset, 1, IRT_EQ, Size_Itemset);
				//rel(*this, Size_Itemset, IRT_GQ, MINSIZE);
				// or 
				linear(*this, Itemset, IRT_GQ, MINSIZE); // minimal size constraint
			}	
			//else
				//linear(*this, Itemset, IRT_GR, 0); // avoid the empty pattern
			
			GCCIM::Closed_Itemset_Constraint(*this, Itemset, MINFREQ);	// the global constraint CLOSEDPATTERN		
										
		};	
		ItemSetClosed(bool share, ItemSetClosed& s) : Script(share, s) {
			Itemset.update(*this, share, s.Itemset);
		}
		virtual Space* copy(bool share) {
			return new ItemSetClosed(share, *this);
		}
		void Branching(){			
			branch(*this, Itemset, INT_VAR_NONE(), INT_VAL_MIN()); // Branching VAR/VAL
		}		
		void Save_Solutions(){ // Save each found solution
			vector <int> v;
			for(int i=0; i<Itemset.size(); i++){
				v.push_back(Itemset[i].val());
			}
			Solutions.push_back(v);
			Frequence_Solution.push_back(CurFreq); // Save frequency of each found solution
		}  
	};
		
	///---- Main() ----///
	int main (int argc, char* argv[])   {
	
		// ------ CPU time Total ------
		struct rusage usage;
		struct timeval end;
		getrusage(RUSAGE_SELF, &usage);	
	
		// Reading Parameters	
		if(argc < 3) {
			cerr << "Error on inputs data : missing parameter(s)" << endl;
			cerr << "Usage : ./main datafile minsup[given in %] minsize[optional]" << endl;
			return 1;
		}				
		string DataBaseFile = argv[1]; // Database file
		double fminsup = atof(argv[2]); // minsup relative (%)	
		int fminsize = 0; // size min (default = 0)
		//int choix_closed_maximal = atoi(argv[3]); // default : activate closed [1 : closed, 2 : maximal]
		if(argc > 3)
			fminsize = atoi(argv[3]); // if minsize is given
		
		// ------ CPU time only Processing database ------
		struct rusage usage0;
		struct timeval end0;
		getrusage(RUSAGE_SELF, &usage0);	
		
		DataBasesInstance.GetData(DataBaseFile); // Load original database
		int minsup = round(fminsup*double(DataBasesInstance.initial_DataBase.size())); // absolute minsup
		DataBasesInstance.Preprocessing_DataBase(minsup); // Processing database (delete infrequents items, encode items) 				
		DataBasesInstance.Create_Boolean_Matrix(); // Generate boolean matrix
		DataBasesInstance.Create_Vertical_DataBase(); // Create vertical representation (bitvector)
		//DataBasesInstance.Create_Horizontal_DataBase(); // Create horizontal representation (bitvector) - for maximal patterns*
		Code = DataBasesInstance.Code; // Get encoding
		Vertical_DataBase = DataBasesInstance.Vertical_DataBase; // Get vertical representation
		//Horizontal_DataBase = DataBasesInstance.Horizontal_DataBase; // Get horizontal representation - for maximal patterns*
		
		// ------ End timer only Processing database ------	
		getrusage(RUSAGE_SELF, &usage0);
		end0 = usage0.ru_utime;
		printf("USER Time (Only pré-traitement) =  %ld.%lds\n", end0.tv_sec, end0.tv_usec);
					
		/// ***** Prints *****
		if(print_){
			// ---- Encoding ---- 
			cout << "--- code -- item ---" << endl;
			for(int g = 0; g < Code.size(); g++){
				cout << g+1 << "--" << Code[g] << endl; 
			}
			// ---- Processed DataBase ---
			vector < vector < int > > N_mat;
			N_mat = DataBasesInstance.Processed_DataBase;
			cout << "--- Processed --- " << endl;
			for(int i = 0; i < N_mat.size(); i++){
				for(int j = 0; j < N_mat[i].size(); j++){
					cout << N_mat[i][j] << " ";
				}cout << endl;
			}		
			// ------ Bool ------
			vector < vector < int > > B_mat;
			B_mat = DataBasesInstance.Boolean_Base;
			cout << "--- Boolean ---" << endl;
			for(int i = 0; i < B_mat.size(); i++){
				for(int j = 0; j < B_mat[i].size(); j++){
					cout << B_mat[i][j] << " ";
				}cout << endl;
			}
			// ------- V --------
			cout << "--- Vertical ---" << endl;
			for(int i = 0; i < Vertical_DataBase.size(); i++){
				for(int j = 0; j < Vertical_DataBase[i].size(); j++){
					cout << Vertical_DataBase[i][j] << " ";
				}cout << endl;
			}
			// ------- H --------
			cout << "--- Horizontal ---" << endl;
			for(int i = 0; i < Horizontal_DataBase.size(); i++){
				for(int j = 0; j < Horizontal_DataBase[i].size(); j++){
					cout << Horizontal_DataBase[i][j] << " ";
				}cout << endl;
			}
			cout << "-----------------------" << endl;
		}
		/// ***** End Prints *****
				
		// ------ CPU time only resolution ------
		float time;
		clock_t t1, t2;		
		t1 = clock();
		
		coverD.Initial_Cover(); // Initial cover (id of all transactions)		
		Resolution(minsup, fminsize); // resolution
				
		// ------ End timer only resolution ------	
		t2 = clock();
		time = (float)(t2-t1)/CLOCKS_PER_SEC;
		printf("CPU Time Only Resolution  = %f\n", time);
		
		// ------ End timer (Total) ------	
		getrusage(RUSAGE_SELF, &usage);
		end = usage.ru_utime;
		printf("USER Time (Total) =  %ld.%lds\n", end.tv_sec, end.tv_usec);			
	
		PostTraitement(Solutions, Code, Frequence_Solution); // Get original patterns according to encoding	
		printStatsGecode(GecodeStat); // Gecode statistics print
		
		return 0;		 		 
	} 
	///---- End Main() ----/// 
	
	///---- Resolution () ----
	void Resolution(int Fmin, int Smin){

		Options opt("/***** Solutions ******/\n");
		opt.solutions(0);
		
		ItemSetClosed* s = new ItemSetClosed(opt, Fmin, Smin);
		s->Branching();
		DFS<ItemSetClosed> e(s); // Depth first search engine
		//Gist::dfs(s);
		delete s;

		while (ItemSetClosed* sol = e.next()){
			sol->Save_Solutions();	
			delete sol;		
		}	
		GecodeStat = e.statistics();
	}
	/// Prints Statistiques Gecode
	void printStatsGecode(const Gecode::Search::Statistics &stats){
		printf("-------------------------------------------------------\n");
		cout <<"Solver search statistics: " << endl;
		cout <<"Propagators executed ================ "<< stats.propagate << endl;
		cout <<"Nodes explored ====================== "<< stats.node << endl;
		cout <<"Failed nodes explored =============== "<< stats.fail << endl;
		cout <<"Maximal depth of explored tree ====== "<< stats.depth << endl;
		cout <<"Peak memory ========================= "<< stats.memory << endl;		
		/* memory ne marche pas sous 4.4.0 */				
	}
	///---- End Resolution () ----///	
	
	///---- PostTraitement () ----
	void PostTraitement(vector < vector <int> > itemset, vector <int> Code, vector <int> Freqs){	
		
		cout << "\nN° of patterns === " << itemset.size() << endl;
		if(print_){
			cout << "\nFrequent Closed Pattrens :\n" << endl;
			// Traduction avec "code"
			vector < vector <int> > Motifs;
			vector <int> motif_temp;
			for (unsigned int i = 0; i < itemset.size(); i++){
				for (unsigned int j = 0; j < itemset[i].size(); j++){				
					if(itemset[i][j] == 1){							
						for (int k =0; k < Code.size(); k++){						
							if (k+1 == j+1){
								motif_temp.push_back(Code[k]);
								break;
							}						
						}
					}
				}
				sort(motif_temp.begin(), motif_temp.end());
				Motifs.push_back(motif_temp);
				// Print closed patterns
				cout << "< ";
				for(unsigned int k = 0; k < motif_temp.size(); k++){
					cout << motif_temp[k] << " ";
				}cout << "> : " << Freqs[i] << endl;
				motif_temp.clear();
			}
		}			
	}
	///---- End PostTraitement () -----///
